unit U_Counttriangles;
{Copyright  2010, Gary Darby,  www.DelphiForFun.org
 This program may be used or modified for any non-commercial purpose
 so long as this original notice remains in place.
 All other rights are reserved
 }

{Program allows users to identify triangles in the displayed figure.  At startup
time it reads arrays of points identifying the figure outline, and the lines which
divide the figure into smaller polygons;  It then identifies all the nodes
(corners) edges (line segments) connecting nodes, and  the "primitive" polygons,
those containing as many nodes and edges.  The primitive polygons are labeled with
single alphabetical letters.  We also identify, but don't immeditaely display,
 all of the triangles which can be formed from the primtive polygons.

The user plays by clicking on primtive polygons, selecting one or more to form a
triangle, then right clicking to add it to their list of found triangles.
Clicking on a previously selected polygon removes it from the trianlge being built.
A message is deliverewd when all traiangles have been found,  A "Hint" button
will provide a clue about one of the triangles not yet found.
The "Show all triangles" button displays a list of all possible triangles.
The "Show more details" button was used primarily for debugging and displays
pixel coordinates for each edge and triangle
}





interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  shellAPI, StdCtrls, ExtCtrls, UGeometry, DFFUtils, ComCtrls ;

type
  TPoly=record
    id:char;  {Identification, one letter for primitive polygons,
                concatenated letters for compound triangles}
    pts:array of TPoint; {coordinates of the polygon corners}
    center:TPoint;  {coordinates of the center of the polygon}
    clearpoint:TPoint; {a point outside of the letter area where we can highlight}
  end;

  TForm1 = class(TForm)
    StaticText1: TStaticText;
    Panel1: TPanel;
    Image1: TImage;
    PageControl1: TPageControl;
    IntroSheet: TTabSheet;
    Memo1: TMemo;
    TriangleSheet: TTabSheet;
    MemoT: TMemo;
    DebugSheet: TTabSheet;
    PageControl2: TPageControl;
    LinesSheet: TTabSheet;
    MemoL: TMemo;
    PrimitivesSheet: TTabSheet;
    Button1: TButton;
    MemoPrims: TMemo;
    ClearBtn: TButton;
    MySheet: TTabSheet;
    ResetBtn: TButton;
    MemoUser: TMemo;
    Edit1: TEdit;
    HintBtn: TButton;
    ShowDetailsBtn: TButton;
    procedure StaticText1Click(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure MemoTClick(Sender: TObject);
    procedure TriangleBtnClick(Sender: TObject);
    procedure MemoLClick(Sender: TObject);
    procedure Image1Click(Sender: TObject);
    procedure Image1MouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure ClearBtnClick(Sender: TObject);
    procedure ResetBtnClick(Sender: TObject);
    procedure HintBtnClick(Sender: TObject);
    procedure ShowDetailsBtnClick(Sender: TObject);
    procedure MemoPrimsClick(Sender: TObject);
    procedure MemoLExit(Sender: TObject);
  public
    nodes:TStringlist;
    edges:TStringList; {define an edge by the index numbers of the 2 end nodes}
    PrimitiveEdges:TStringList;
    PrimitivePolys:array of TPoly;
    Allpolys:TStringlist;
    Usertriangles:TStringlist;
    lastline:TLine; {last line highlighted, so it can be un-lighted before drawing next}
    highlightIds:string; {Letter ids of polygons which are highlighted}
    hintcount:integer;
    procedure Initializecase;
    function NodeKeyFromPoint(p:TPoint):string;
    function PointFromNodekey(key:string):TPoint;
    function EdgeKeyFromPoints(p1,p2:TPoint):string;
    procedure drawnode(p:TPoint);
    function Connected(p1,p2:TPoint):boolean;
    function PrimitiveConnected(p1,p2:TPoint):boolean;
    function Colinear(p1,p2,p3:TPoint):boolean;
    procedure drawline(p1,p2:TPoint);
    procedure drawedge(key:string);
    procedure LabelprimitiveTriangles;
    procedure Togglehighlight(P:TPoint); {Highight or "Unhighlight" the primitive poly at P}
    procedure Highlight(P:TPoint;index:integer);{Highight the primitive poly at P}
    {"UnHighight" the primitive poly at P}
    procedure Unhighlight(P:TPoint;Index:integer); overload;
    procedure Unhighlight(CheckId:char); overload;
    procedure makealltriangles;
  end;

var
  Form1: TForm1;

implementation

uses math;

{$R *.DFM}

var
  PatOutline:array [0..7] of TPoint= ((x:1;y:0), (x:4;y:0),(x:5;y:1),(x:5;y:4),
                                    (x:4;y:5), (x:1;y:5),(x:0; y:4),(x:0;y:1));

  {The interior pattern lines running from one point ot he boundary to another}
  {the values here are index vale into the PatOutlines array}
  Patlines:array [0..7] of TPoint= ((x:0;y:6), (x:0;y:5),(x:0;y:4),(x:1;y:5),
                                    (x:1;y:4), (x:1;y:3),(x:2; y:6),(x:3;y:7));

  TwoPi:extended=2*Pi;

  LetterOffsetX, LetterOffsetY:integer;

{************* FormActivate **********}
procedure TForm1.FormActivate(Sender: TObject);
var
  i:integer;
  scale,figscale:real;
  minx,miny,maxx,maxy:real;
  offsetx,offsety:integer;
begin

  {Initialize stringlists, etc.}
  nodes:=TStringList.create;
  nodes.Sorted :=true;
  edges:=TStringList.create;
  edges.Sorted:=true;
  PrimitiveEdges:=TStringList.create;
  PrimitiveEdges.Sorted:=true;
  Allpolys:=TStringlist.create;
  Allpolys.sorted:=true;
  UserTriangles:=TStringlist.Create;
  Usertriangles.Sorted:=true;

  {Find scaling factos}
  with Image1 , canvas do
  begin
    minx:=9999; miny:=9999;
    maxx:=0; maxy:=0;
    for i:=0 to  high(PatOutline) do
    with PatOutline[i] do
    begin
      maxx:=max(maxx,x);
      minx:=min(minx,x);
      maxy:=max(maxy,y);
      miny:=min(miny,y);
    end;
    figscale:= max(maxx-minx,maxy-miny);
    scale:= 9*min(width,height) / (10*figscale);
    offsetx:=width div 20;
    offsety:=height div 20;

    {Identify all nodes}
    for i := 0 to high(PatOutline) do
    with PatOutline[i]  do
    begin
      x:=offsetx+round(x*scale);
      y:=offsety+round(y*scale);
    end;

    {draw the outline and the interiot lines}
    polygon(PatOutline);
    for i:=0 to high(Patlines) do
    with Patlines[i] do
    begin
      with PatOutline[x] do moveto(x,y);
      with PatOutline[y] do lineto(x,y);
    end;
    {Offsets to center label letters at polygon centroid}
    letteroffsetx:=textwidth('X') div 2;
    letteroffsetY:=textheight('Y') div 2;
  end;
  InitializeCase;
end;

{************* DrawNode ***********}
procedure TForm1.DrawNode(p:TPoint);
{Draw a blue circle at this point}
begin
  with Image1 , canvas, p do
  begin
    brush.color:=clblue;
    ellipse(x-4, y-4, x+4, y+4);
  end;
end;

{************** DrawLine ***********}
procedure TForm1.drawline(p1,p2:TPoint);
{redraw any old line in black and  draw the new line in red}
begin
  with image1, canvas do
  begin
    if lastline.p1.x>=0 then
    with lastline do
    begin
      pen.color:=clBlack;
      moveto(p1.x, p1.y);
      lineto(p2.x, p2.y);
    end;
    pen.color:=clred;
    moveto(p1.x, p1.y);
    lineto(p2.x, p2.y);
  end;
  lastline:=line(p1,p2);
end;

{*************** DrawEdge **********}
procedure TForm1.drawedge(key:string);
{Given 12 character [(3,3),(3,3)] line definition string, draw it}
var
  p1,p2:TPoint;
begin
  p1.x:=strtoint(copy(key,1,3));
  p1.y:=strtoint(copy(key,4,3));
  p2.x:=strtoint(copy(key,7,3));
  p2.y:=strtoint(copy(key,10,3));
  drawline(p1,p2);
end;

{***************** NodekeyFromPoint ***********}
function TForm1.NodekeyFromPoint(p:TPoint):string;
begin
  result:=format('%3d%3d',[(p.x),(p.y)]);
end;

{*********** PointFromNodeKey **************}
function TForm1.PointFromNodekey(key:string):TPoint;
{Reverse the string built by NodeKeyFromPoint to return a TPoint record}
begin
  result.x:=strtoint(copy(key,1,3));
  result.y:=strtoint(copy(key,4,3));
end;

{************* EdgekeyFromPoints ************}
function TForm1.edgekeyFromPoints(p1,p2:TPoint):string;
{Make a 12 character string representing an edge line from 2 points}
{Sorted so that furthest left and top point is first}
begin
  if (p1.x<p2.x) or ((p1.x=p2.x) and (p1.y<p2.y))
  then result:=format('%3d%3d%3d%3d',[p1.x,p1.y,p2.x,p2.y])
  else result:=format('%3d%3d%3d%3d',[p2.x,p2.y,p1.x,p1.y]);
end;

{************ Colinear **********8}
function TForm1.Colinear(p1,p2,p3:TPoint):boolean;
{The two shorter sides of a triangle must exceed the length of the third side}
{Here we require that the sum be at least 3 pixels greater than the longest side
 to allow for rounding errors caused by using integer coordinates}
{Colinear will return True of False based on the results of the test}
var
  s1,s2,s3:integer;

  function dist(p1,p2:TPoint):integer;
  {distance between two pints}
  begin
    result:=round(sqrt(sqr(p1.x-p2.x)+sqr(p1.y-p2.y)));
  end;

  procedure swap(var a,b:integer);
  {swap two integers}
  var temp:integer;
  begin
    temp:=a;
    a:=b;
    b:=temp;
  end;

begin
  s1:=dist(p1,p2);  {Get the side lengths}
  s2:=dist(p2,p3);
  s3:=dist(p1,p3);
  if s1>s2 then swap(s1,s2);  {Sort side by increasing length}
  if s2>s3 then swap(s2,s3);
  if s1+s2>=s3+3 then result:=false else result:=true;;
end;



{************** Connected **************}
function TForm1.Connected(p1,p2:TPoint):boolean;
{Is there a line segment connecting these two points?}
var
  s:string;
  index:integer;
begin
  s:=edgekeyFromPoints(p1,p2);
  if edges.find(s,index)
  then result:=true else result:=false;
end;

{************** PrimitiveConnected **************}
function TForm1.PrimitiveConnected(p1,p2:TPoint):boolean;
{Is there a line segment connecting these two points?}
var
  s:string;
  index:integer;
begin
  s:=edgekeyFromPoints(p1,p2);
  if PrimitiveEdges.find(s,index)
  then result:=true else result:=false;
end;

{************* InitializeCase ***********8}
procedure TForm1.initializecase;
var
  i,j,k,index:integer;
  L1,L2:TLine;
  OnBorder:boolean;
  IP:TPoint;
  s:string;
  SubNodes:TStringList;
  prevs:string;
  nodedir, prevnodedir:string;
  p1,p2:TPoint;
  angle,angle2:extended;

  function dist(s:string):integer;
  var x1,x2,y1,y2:integer;
  begin
    x1:=strtoint(copy(s,1,3));
    y1:=strtoint(copy(s,4,3));
    x2:=strtoint(copy(s,7,3));
    y2:=strtoint(copy(s,10,3));
    result:=trunc(sqrt(sqr(x1-x2)+sqr(y1-y2)));
  end;

begin
  SubNodes:=TStringList.Create;
  {find all the line intersections}
  lastline.p1.x:=-1;
  for i:= 0 to high(PatOutline) do
  begin
    nodes.Add(NodekeyFromPoint(patoutline[i]));
    drawnode(patoutline[i]);
    {Save the line segments (edges) connecting the intersections (nodes)}
    If i>0 then
    begin
      s:=edgekeyfrompoints(patoutline[i-1],patoutline[i]);
      if not edges.Find(s,index) then edges.add(s);
    end
    else
    begin
      s:=edgekeyfrompoints(patoutline[0],patoutline[high(patoutline)]);
      if not edges.Find(s,index) then edges.add(s);
    end;
  end;


  for i:= 0 to high(Patlines)  do
  with Patlines[i] do
  begin   {check every pattern line to find its intersections}
    L1:=Line(PatOutline[x],patoutline[y]);
    subnodes.clear;
    {intersections start with the first border point}
    s:=NodeKeyfrompoint(Patoutline[x]);
    subnodes.add(s);

    for j:=0 to high(Patlines) do
    if i<>j  then {for all other pattern lines}
    begin {looking for an intersection}
      with Patlines[j] do L2:=Line(PatOutline[x],patoutline[y]);
      if intersect(L1,L2,OnBorder,IP) then  {Intersect checking function}
      begin {the 2 lines do intersect}
        s:=nodekeyfromPoint(IP); {define the new node}
        subnodes.add(s); {and add it with distance to the intersections list}
        if not nodes.Find(s,index) then
        begin {If it's new, add it to the list of all nodes}
          nodes.Add(s);
          DrawNode(IP);
        end;
      end;
    end;
    {we have all of the interior intersections on line i, add the end border point}
    subnodes.add(nodekeyfrompoint(patoutline[y]));
    {make line segments for all intersection point to intersection point
     connections along line i}
    for j:=0 to subnodes.Count-2 do
    begin  {for each node on line i}
      for k:=j+1 to subnodes.count-1 do
      if subnodes[j]<>subnodes[k] then
      begin  {define an edge to every other node on line i}
        s:=EdgeKeyfromPoints(PointFromNodeKey(subnodes[j]), PointFromNodeKey(subnodes[k]));
        if not edges.find(s,index) then edges.Add(s);
      end;
    end;
  end;

  {Display the line segments}
  MemoL.clear;
  memoL.lines.add(format('Pixel coordinates for the %d line segments found',
                         [edges.count]));
  memoL.lines.Add('Click any line to view that segment');
  subnodes.Clear;
  subnodes.sorted:=true;
  for i:=0 to edges.count-1 do
  begin
    s:=edges[i];
    memoL.lines.add(format('#%2d:(%3s,%3s) to (%3s,%3s)',
              [i+1, copy(s,1,3),copy(s,4,3),copy(s,7,3),copy(s,10,3)]));
    {filter to find primitive edges, those directly connected to the next node}
    {for each line generate a record for each line and with key
     containing Node IDs, the angle (0 to 360)  and length of the line.
     when sorted, the first of those with identical directions will be the shortes, therefore
     the desirec primitive edge }
      p1:=pointfromnodekey(copy(s,1,6));
      p2:=pointfromnodekey(copy(s,7,6));
      angle := arctan2(p1.y-p2.Y,p1.x-p2.x);
      angle2:=angle+pi;
      if angle<0 then angle:=angle+twopi;

      if angle2>twopi then angle2:=angle2-twopi;
      subnodes.add(copy(s,1,6)+format('%3.3d%3.3d',[trunc(angle*60/pi),dist(s)])+copy(s,7,6));
      subnodes.add(copy(s,7,6)+format('%3.3d%3.3d',[trunc(angle2*60/pi),dist(s)])+copy(s,1,6));
  end;
  subnodes.sort;
  prevnodedir:='';
  primitiveEdges.clear;
  for i:=0 to subnodes.count-1 do
  begin
    s:=subnodes[i];
    nodedir:=copy(s,1,9);
    if copy(s,1,9)<>prevnodedir then
    begin
      p1:=pointfromNodekey(copy(s,1,6));
      p2:=pointfromnodekey(copy(s,13,6));
      s:=Edgekeyfrompoints(p1,p2);
      primitiveEdges.add(s);
      prevnodedir:=nodedir;
    end;
  end;
  memoL.lines.add(#13+'Primitive Lines');
  prevS:='';
  for i:=0 to primitiveEdges.count-1 do
  begin
    s:=copy(PrimitiveEdges[i],1,12);
    if s<>prevS then
    begin
      memoL.lines.add(format('#%2d:(%3s,%3s) to (%3s,%3s)',
              [i+1, copy(s,1,3),copy(s,4,3),copy(s,7,3),copy(s,10,3)]));

      prevs:=s;
    end;
  end;
  subnodes.free;
  movetotop(memoL);
  labelPrimitiveTriangles;
  HighlightIds:='';
  makealltriangles;
  Pagecontrol1.ActivePage:=Introsheet;
end;

procedure Tform1.LabelprimitiveTriangles;
var
  polycount:integer;
  i,j,k,L,index:integer;
  p:array[0..3] of TPoint;
  dupchecklist:TStringlist;
  s:string;

     function sortnodes(p:array of TPoint;n:integer):string;
     var
       i,j:integer;
       temp:TPoint;
     begin
       for i:=0 to n-2 do
       for j:=i+1 to n-1 do
       if (p[i].x>p[j].x) or
          ((p[i].x=p[j].x) and (p[i].y >=p[j].y)) then
       begin
         temp:=p[i];
         p[i]:=p[j];
         p[j]:=temp;
       end;
       result:='';
       for  i:=0 to  n-1 do result:=result+nodekeyfrompoint(p[i]);
     end;

    procedure checkpoly(p:array of tPoint; N:integer);
    var
      i:integer;
      s2:string;
      cx,cy:integer;
      ch:char;
      m:integer;
    begin
        s2:=sortnodes(p,n);
        if not dupcheckList.find(s2,index) then
        begin
          dupchecklist.add(s2);
          inc(polycount);
          cx:=0;
          cy:=0;
          for i:=0 to n-1 do
          begin
            inc(cx,p[i].X);
            inc(cy,p[i].Y);
          end;
          cx:=cx div n;
          cy:=cy div n;
          image1.canvas.brush.color:=clwhite;
          ch:= char(ord('A')-1+polycount);
          image1.Canvas.TextOut(cx-letteroffsetx, cy-letteroffsetY, ch );
          s:='';
          for m:=0 to n-1 do s:=s+format('(%3d,%3d) ',[p[m].X,p[m].y ]);
          MemoPrims.lines.add(format('%s: %s)',[ch,s]));
          setlength(primitivepolys,polycount);
          with PrimitivePolys[polycount-1] do
          begin
            id:=ch;
            setlength(pts,n);
            for m:=0 to n-1 do pts[m]:=p[m];
            center:=point(cx,cy);
            {define a safe point to simulate a click for highlighting}
            clearpoint.x:= (p[0].X+cx) div 2;
            clearpoint.Y:=(p[0].Y + cy) div 2;
          end;
        end;
      end;


begin
  dupchecklist:=TStringlist.create;
  dupchecklist.sorted:=true;
  polycount:=0;
  setlength(primitivepolys,0);
  memoprims.Clear;
  MemoPrims.lines.add('Pixel coordinates for primitivePolygons');
  MemoPrims.lines.add('Click any line to view that Polygon');

  for i:= 0 to nodes.count-1 do
  begin
    p[0]:=PointfromNodekey(nodes[i]);
    for j:= 0 to nodes.count-1 do
    if i<>j then
    begin
      p[1]:=PointfromNodekey(nodes[j]);
      if primitiveconnected(p[0],p[1]) then
      for k:= 0 to nodes.count-1 do
      if (k<>i) and (k<>j) then
      begin
        p[2]:=PointfromNodekey(nodes[k]);
        if primitiveconnected(p[0],p[2]) and primitiveconnected (p[1],p[2])
          and (not Colinear(p[0],p[1],p[2]))
        then  checkpoly(p,3) {found a triangle}
        else for L:=0 to nodes.count-1 do
        if (L<>I) and (L<>j) and (L<>K)  then
        begin     {check for quadralaterals}
          p[3].x:=strtoint(copy(nodes[L],1,3));
          p[3].y:=strtoint(copy(nodes[L],4,3));
          if primitiveconnected(p[0],p[1]) and primitiveconnected (p[1],p[2])
          and primitiveconnected(p[2],p[3]) and primitiveconnected(p[3],p[0])
          and (not primitiveconnected(p[0],p[2])) and (not primitiveconnected(p[1],p[3]))
          and (not Colinear(p[0],p[1],p[2])) and (not (Colinear(p[1],p[2],p[3]))
          and (not Colinear(p[0],p[2],p[3])))
          then  checkpoly(p,4);
        end;
      end;
    end;
  end;
  dupchecklist.free;
  MemoPrims.lines[0]:=format('Pixel coordinates for the %d primitive Polygons',
                           [polycount]);
end;

{************** TriangleBtnClick **********}
procedure TForm1.TriangleBtnClick(Sender: TObject);
var i:integer;
begin
  MemoT.Clear;
  MemoT.lines.add(format('%d triangles found',[allpolys.Count]));
  MemoT.lines.add('Click any line to view that triangle');
  for i:=0 to allpolys.count-1 do
     MemoT.lines.add(format('#%2d: %s',[i+1, allpolys[i]]));
  movetotop(memot);
  Trianglesheet.TabVisible:=true;
  pagecontrol1.activepage:=TriangleSheet;
end;


{************** MakeAllTriangles ************}
procedure TForm1.makealltriangles;
{Find all triangles}
var
  i,j,k,m:integer;
  key:string;
  p:array[0..2] of TPoint;
begin
  {Search for triangles by check all triples of nodes for
   being connected and not  colinear}
  for i:= 0 to nodes.count-3 do
  begin
    p[0].x:=strtoint(copy(nodes[i],1,3));
    p[0].y:=strtoint(copy(nodes[i],4,3));
    for j:= i+1 to nodes.count-2 do
    begin
      p[1].x:=strtoint(copy(nodes[j],1,3));
      p[1].y:=strtoint(copy(nodes[j],4,3));
      if connected(p[0],p[1]) then
      for k:= j+1 to nodes.count-1 do
      begin
        p[2].x:=strtoint(copy(nodes[k],1,3));
        p[2].y:=strtoint(copy(nodes[k],4,3));
        if connected(p[0],p[2]) and connected (p[1],p[2])
          and (not Colinear(p[0],p[1],p[2]))
        then
        begin  {found a triangle}
          {find the letters representing the primitive polygons included}
          {Method, find which primitive polygons are cointained in this triangle
          and construct a key the their letter ids for the Allpolys list}
          key:='';
          for m:=0  to high(primitivepolys) do
          with primitivepolys[m] do
          if pointinpoly(center,p)=ppinside then key:=key+id;
           allpolys.add(key);
        end;
      end;
    end;
  end;
end;

{************* MemoLClick ***********}
procedure TForm1.MemoLClick(Sender: TObject);
{Lines memo clicked to display an line segment in red}
var
  n,index:integer;
  s,key:string;
begin
  n:=memoLineclicked(memoL);
  s:=memoL.lines[n];
  if s[1]='#' then
  begin
    key:=copy(s,6,3)+copy(s,10,3)+copy(s,19,3)+copy(s,23,3);
    if edges.find(key,index) then drawedge(key);
  end;
end;

{********* Memo1Exit *************}
procedure TForm1.MemoLExit(Sender: TObject);
{leaving MemoL, the edges display memo}
begin
  {in case there is a red line showing from user clicks, we need to redraw
  it in black}
  with Image1,canvas do
  begin
     if lastline.p1.x>=0 then
    with lastline do
    begin
      pen.color:=clBlack;
      moveto(p1.x, p1.y);
      lineto(p2.x, p2.y);
    end;
  end;
end;


{************ MemoTClick ************}
procedure TForm1.MemoTClick(Sender: TObject);
{Click on all triangles memo line to highlight it}
var
  i,n,index:integer;
  s:string;
  key:string;
  memo:TMemo;
begin
  memo:=TMemo(sender);
  n:=memoLineclicked(Memo);
  s:=Memo.lines[n];
  if length(s)=0 then exit;
  if s[1]='#' then {user wants to see a triangle}
  begin
    {Unhighlight any already highlighted}
    for i:=0 to high(primitivepolys) do
    with primitivepolys[i] do  unhighlight(clearpoint,i);
    key:=copy(s,6,length(s)-5);
    for i:=1 to length(key) do
    begin  {for each primitive }
      index:=ord(key[i])-ord('A');
      with primitivepolys[index] do
      begin
        highlight(clearpoint,index);
      end;
    end;
  end;
end;

{*************** MemoPrimsClick *********}
procedure TForm1.MemoPrimsClick(Sender: TObject);
{Display the Primitive polygon line clicked by the user}
var
  i,n,index:integer;
  s:string;
begin
  {unhighlight all}
  for i:=0 to high(primitivepolys) do
    with primitivepolys[i] do  unhighlight(clearpoint,i);
  with memoprims do
  begin
    n:=ClickedMemoLine(MemoPrims);
    s:=lines[n];
    if (length(s)>4) and (s[2]=':') then
    begin
      index:=ord(s[1])-ord('A');
      with primitivepolys[index] do  highlight(clearpoint,index);
    end;
  end;
end;

{************** Highlight ***********}
procedure TForm1.Highlight(P:TPoint;index:integer);
{Select and Highlight a primitive polygon}
var
  ClickId:char;
  fillpoint:TPoint;
begin
  with primitivepolys[index] do
  begin
    Clickid:=Id;
    {In case the user clicked on the label letter, move the floodfill point a
     little.  Floodfill interior to a letter (e,g "B" or "O", etc.) will  only
     fill the interior part  of the letter, not the entire polygon}
     fillpoint:=clearpoint;
   end;
  with image1, canvas, fillpoint do
  begin
    brush.color:=clyellow;
    floodfill(x,y,clblack, fsborder);
    update;
  end;
  Highlightids:=highlightids+Clickid;
  edit1.Text:=highlightids;
end;

{**************** Unhighlight **************}
procedure TForm1.UnHighlight(P:TPoint;Index:integer);
{Deselect a selected primitve polygon}
  var
  ClickId:char;
  fillpoint:TPoint;
begin
  with primitivepolys[index] do
  begin
    Clickid:=Id;
    {In case the user clicked on the label letter, move the floodfill point a
     little.  Floodfill interior to a letter (e,g "B" or "O", etc.) will  only
     fill the interior part  of the letter, not the entire polygon}
     fillpoint:=clearpoint;
   end;
  with image1, canvas, fillpoint do
  begin
    brush.color:=clWhite;
    Canvas.floodfill(x,y,clblack, fsborder);
    update;
  end;
  highlightIds:=stringreplace(highlightids,ClickId,'',[rfreplaceall]);
  edit1.Text:=highlightids;
end;

{**************** Unhighlight **************}
procedure TForm1.UnHighlight(CheckID:char);
{Override unhighlight version from polygon Id only}
var
  p:TPoint;
  i:integer;
begin
  for i:=0 to high (primitivePolys) do
  with primitivePolys[i] do
  if Checkid=id then
  begin
    p:=clearpoint;
    with image1, canvas do
    begin
      brush.color:=clWhite;
      Canvas.floodfill(p.x,p.y,clblack, fsborder);
    end;
    highlightIds:=stringreplace(highlightids,Id,'',[rfreplaceall]);
    edit1.Text:=highlightids;
  end;
end;

{************ ToggleHighlight *************}
procedure TForm1.ToggleHighlight(P:TPoint);
{Highlight or "unhighlight" based on current state}
var
  i:integer;
begin
  with image1,canvas, p do
  begin
    for i:=0  to high(primitivepolys) do
    with primitivepolys[i] do
    if pointinpoly(p,pts)=ppinside then

    begin
      if pixels[clearpoint.x,clearpoint.y]=clyellow
      then UnHighlight(P,i) else Highlight(p,i);
      break;
    end;
  end;
end;

{************** Image1Click **********}
procedure TForm1.Image1Click(Sender: TObject);
{Select or deselect a polygon}
var
  p:TPoint;
begin
  pagecontrol1.ActivePage:=MySheet;
  p:=Image1.screentoclient(mouse.cursorpos);
  TogglehighLight(p);
end;

{*************** Image1MouseUp *************}
procedure TForm1.Image1MouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
{Right buttom clicks will analyze selected tpolygons to see if they
 form a triangle to be added to the users collection}
var
  i,j:integer;
  s:string;
  temp:char;
  index:integer;
  msg:string;  {COngratulations message when solved}
begin

  if button = mbright then
  begin     {check for triangle}
    s:=highlightids;
    if length(s)>1 then
    for i:=1 to length(s)-1 do
    begin
      for j:=i+1 to length(s) do
      if s[i]>s[j] then
      begin
        temp:=s[i];
        s[i]:=s[j];
        s[j]:=temp;
      end;
    end;
    if allpolys.Find(s,j) then
    with memouser do
    begin {it is a triangle}
      if not usertriangles.find(s,index) then
      begin
        usertriangles.Add(s);
        lines.add(format('#%2d: %s',[usertriangles.count,s]));
        for i:=1 to length(s) do unhighlight(s[i]);
        if userTriangles.count=allpolys.Count
        then
        begin
          if hintcount=0
          then msg:='Congratulations, you have found all triangles on your own!!!'
          else
          case (hintcount-1) div 3 of
            0: {1,2,3 hints}
              msg:='Congratulations, you found all triangles with only a little help from me!!';
            1: {4,5,6 hints}
              msg:='You found all the triangles but you can do it with less help!';
            2: {7,8,9 hints}
              msg:='Together, we found all the triangles but I''m sure you can do it with less help.';
            else
              msg:='Come on now, you''re not trying hard enough. I''m sure you can do it with less help.';
          end; {case}
          showmessage(msg);
        end;
      end
      else beep;
    end;
  end;
end;

{************ ClearBtnClick ************}
procedure TForm1.ClearBtnClick(Sender: TObject);
var
  i:integer;
begin
  for i:= 0 to high(Primitivepolys) do
  with primitivepolys[i] do unhighlight(clearpoint,i);
  highlightIds:='';
end;

{************* ResetBtnClick *********}
procedure TForm1.ResetBtnClick(Sender: TObject);
{Reset everything that might have change during play}
begin
  memouser.Clear;
  usertriangles.Clear;
  clearbtnclick(sender);
  hintcount:=0;
  TriangleSheet.Tabvisible:=false;
  DebugSheet.TabVisible:=false;
  Pagecontrol1.activepage:=Introsheet;
end;


{************* HintBtnClick **********8}
procedure TForm1.HintBtnClick(Sender: TObject);
{User wants a hint, find all triangled not in his list and display the
 1 character and the number of primitives for one of them selected randomly}
var
  i,index:integer;
  templist:TStringlist;
  hint:string;
begin
  Templist:=tstringlist.create;
  for i:=0 to allpolys.Count-1 do
  begin
     index:=userTriangles.IndexOf(allpolys[i]);
     if index<0 then templist.add(allpolys[i]);
  end;
  if templist.count>0 then
  begin
    hint:=templist[random(templist.count)];
    showmessage(format('Find one with %d primitive triangles, one of which is %s',
                               [length(hint),hint[1]]));
    inc(hintcount);
  end;
  templist.free;
end;


{********** ShowDetailsBtnClick ********}
procedure TForm1.ShowDetailsBtnClick(Sender: TObject);
begin
  DebugSheet.TabVisible:=true;
  PageControl1.activepage:=Debugsheet;
end;

procedure TForm1.StaticText1Click(Sender: TObject);
begin
  ShellExecute(Handle, 'open', 'http://www.delphiforfun.org/',
  nil, nil, SW_SHOWNORMAL) ;
end;

end.
